using System;
using System.IO;
using System.Text;

namespace AmortizationCalculator
{
    public class ExcelSpreadsheetGenerator
    {
        private AmortizationModel _model;

        #region constants
        string HEADER1 = @"<?xml version=""1.0""?>
<?mso-application progid=""Excel.Sheet""?>
<Workbook xmlns=""urn:schemas-microsoft-com:office:spreadsheet"" xmlns:o=""urn:schemas-microsoft-com:office:office"" xmlns:x=""urn:schemas-microsoft-com:office:excel"" xmlns:ss=""urn:schemas-microsoft-com:office:spreadsheet"" xmlns:html=""http://www.w3.org/TR/REC-html40"">
 <Styles>
  <Style ss:ID=""s21"">
   <NumberFormat ss:Format=""Currency""/>
  </Style>
  <Style ss:ID=""s22"">
   <NumberFormat ss:Format=""Short Date""/>
  </Style>
  <Style ss:ID=""s23"">
   <Alignment ss:Horizontal=""Center"" ss:Vertical=""Bottom""/>
   <Font x:Family=""Swiss"" ss:Bold=""1""/>
  </Style>
 </Styles>
 <Worksheet ss:Name=""Amortization Schedule"">
  <Table ss:ExpandedColumnCount=""8"" ss:ExpandedRowCount=""";
        string HEADER2 = @""" x:FullColumns=""1"" x:FullRows=""1"">
   <Column ss:Width=""98.25""/>
   <Column ss:Width=""144.75""/>
   <Column ss:Index=""5"" ss:Width=""75.75""/>
   <Column ss:Index=""7"" ss:Width=""54.75""/>
   <Column ss:Width=""72.75""/>";
        string FOOTER = @"  </Table>
 </Worksheet>
</Workbook>";
        string StyleCurrency = "s21";
        string StyleDate = "s22";
        string StyleBold = "s23";
        #endregion

        public ExcelSpreadsheetGenerator(AmortizationModel model)
        {
            _model = model;
        }
        private FileStream _stream;
        private TextWriter _writer;
        public void CreateFile(string filename)
        {
            if (filename == null)
                return;
            _stream = null;
            _writer = null;
            try
            {
                _stream = File.Create(filename);
                _writer = new StreamWriter(_stream);

                int qtyRows = 15 + _model.Count;
                _writer.WriteLine(HEADER1 + qtyRows.ToString() + HEADER2);
                addRow(0,
                    createCell(null, "String", "Principal"),
                    createCell(StyleCurrency, "Number", formatCurrency(_model.Principal)));
                addRow(0,
                    createCell(null, "String", "APR"),
                    createCell(null, "Number", _model.Apr.ToString()));
                addRow(0,
                    createCell(null, "String", "Start Date"),
                    createCell(null, "String", _model.StartDate.ToLongDateString()));

                addRow(5,
                    createCell(null, "String", "Payment Frequency"),
                    createCell(null, "String", _model.PaymentsPerYear.ToString()));
                addRow(0,
                    createCell(null, "String", "Number of Payments"),
                    createCell(null, "Number", _model.QtyPayments.ToString()));
                addRow(0,
                    createCell(null, "String", "Number of Years"),
                    // not really currency but formatted with the same rules
                    createCell(null, "Number", formatCurrency(_model.QtyYears)));

                addRow(9,
                    createCell(null, "String", "Payment"),
                    createCell(StyleCurrency, "Number", formatCurrency(_model.PaymentAmount)));
                addRow(0,
                    createCell(null, "String", "Number of Payments"),
                    createCell(null, "Number", _model.QtyPaymentsWithExtraPayments.ToString()));
                addRow(0,
                    createCell(null, "String", "Total Repaid"),
                    createCell(StyleCurrency, "Number", formatCurrency(_model.TotalRepaid)));
                addRow(0,
                    createCell(null, "String", "Total Interest Paid"),
                    createCell(StyleCurrency, "Number", formatCurrency(_model.TotalInterestPaid)));
                addRow(0,
                    createCell(null, "String", "Total Principal Paid"),
                    createCell(StyleCurrency, "Number", formatCurrency(_model.TotalPrincipalPaid)));

                addRow(15,
                    createCell(StyleBold, "String", "#"),
                    createCell(StyleBold, "String", "Date"),
                    createCell(StyleBold, "String", "Interest"),
                    createCell(StyleBold, "String", "Principal"),
                    createCell(StyleBold, "String", "Extra Payment"),
                    createCell(StyleBold, "String", "Cum Int"),
                    createCell(StyleBold, "String", "Cum Princ"),
                    createCell(StyleBold, "String", "Princ Balance"));

                foreach (SinglePayment payment in _model)
                {
                    addRow(0,
                        createCell(null, "Number", payment.PaymentNumber.ToString()),
                        createCell(StyleDate, "DateTime", String.Format("{0:yyyy-MM-ddT00:00:00.000}", payment.PaymentDate)),
                        createCell(StyleCurrency, "Number", formatCurrency(payment.Interest)),
                        createCell(StyleCurrency, "Number", formatCurrency(payment.Principal)),
                        createCell(StyleCurrency, "Number", formatCurrency(payment.ExtraPayment)),
                        createCell(StyleCurrency, "Number", formatCurrency(payment.CumulativeInterest)),
                        createCell(StyleCurrency, "Number", formatCurrency(payment.CumulativePrincipal)),
                        createCell(StyleCurrency, "Number", formatCurrency(payment.PrincipalAtEndOfPeriod))
                        );
                }

                _writer.WriteLine(FOOTER);
            }
            catch { }
            finally
            {
                if (_writer != null)
                    _writer.Close();
                if (_stream != null)
                    _stream.Close();
            }
        }
        private string createCell(string style, string dataType, string text)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("<Cell");
            sb.Append((String.IsNullOrEmpty(style)) ? ">" : @" ss:StyleID=""" + style + @""">");
            sb.Append(@"<Data ss:Type=""" + dataType + @""">");
            sb.Append(text);
            sb.Append("</Data></Cell>");

            return sb.ToString();
        }

        private void addRow(int index, params string[] cells)
        {
            _writer.WriteLine("   <Row" + ((index > 0) ? @" ss:Index=""" + index.ToString() + @"""" : "") + ">");
            foreach (string cell in cells)
                _writer.WriteLine("    " + cell);
            _writer.WriteLine("   </Row>");
        }

        private string formatCurrency(double currency)
        {
            // don't combine, or it could trim zeros to the left of the decimal as well
            string retStr = String.Format("{0:0.00}", currency).TrimEnd('0');
            retStr = retStr.TrimEnd('.');
            return retStr;
        }
    }
}
// 