using System;
using System.Collections;
using System.Collections.Generic;

namespace AmortizationCalculator
{
    public enum PaymentsPeriod { Annually = 1, SemiAnnually = 2,
        EveryFourMonths = 3, Quarterly = 4, BiMonthly = 6, Monthly = 12 }

    public class AmortizationModel : IEnumerable
    {
        //
        // set by user
        //
        private double _principal = 0;
        private double _apr = 0;
        private PaymentsPeriod _paymentsPerYear;
        private int _qtyPayments = 0;
        private DateTime _startDate = DateTime.Now;

        private double _qtyYears;

        //
        // calculated
        //
        private double _paymentAmount;
        private double _totalInterestPaid;
        private double _totalPrincipalPaid;
        private double _totalRepaid;
        private int _qtyPaymentsWithExtraPayments;

        private int _monthsBetweenPayments;

        private AbbrevList<SinglePayment> _payments = new AbbrevList<SinglePayment>();

        #region Public properties, set by user
        /// <summary>Get or set</summary>
        public double Principal
        {
            get { return _principal; }
            set
            {
                if (_principal == value)
                    return;
                _principal = value;
                OnLoanTermsChanged();
                calculate();
            }
        }
        /// <summary>Get or set</summary>
        public double Apr
        {
            get { return _apr * 100; }
            set
            {
                double alteredValue = value / 100;
                if (_apr == alteredValue)
                    return;
                _apr = alteredValue;
                OnLoanTermsChanged();
                calculate();
            }
        }
        /// <summary>Get or set</summary>
        public PaymentsPeriod PaymentsPerYear
        {
            get { return _paymentsPerYear; }
            set
            {
                if (_paymentsPerYear == value)
                    return;
                _paymentsPerYear = value;
                OnLoanTermsChanged();
                calculate();
            }
        }
        /// <summary>Get or set</summary>
        public int QtyPayments
        {
            get { return _qtyPayments; }
            set
            {
                if (_qtyPayments == value)
                    return;
                _qtyPayments = value;
                OnLoanTermsChanged();
                calculate();
            }
        }
        /// <summary>Get or set</summary>
        public DateTime StartDate
        {
            get { return _startDate; }
            set
            {
                if (_startDate == value)
                    return;
                _startDate = value;
                OnLoanTermsChanged();
                calculate();
            }
        }
        #endregion
        #region Public Properties, calculated
        /// <summary>Get</summary>
        public double QtyYears
        {
            get { return _qtyYears; }
        }
        /// <summary>Get</summary>
        public double PaymentAmount
        {
            get { return _paymentAmount; }
        }
        /// <summary>Get</summary>
        public double TotalInterestPaid
        {
            get { return _totalInterestPaid; }
        }
        /// <summary>Get</summary>
        public double TotalPrincipalPaid
        {
            get { return _totalPrincipalPaid; }
        }
        /// <summary>Get</summary>
        public double TotalRepaid
        {
            get { return _totalRepaid; }
        }
        /// <summary>Get</summary>
        public int QtyPaymentsWithExtraPayments
        {
            get { return _qtyPaymentsWithExtraPayments; }
        }
        #endregion
        #region indexer,count, enumerator
        /// <summary>indexer. accesses payment lines</summary>
        /// <param name="index"></param>
        /// <returns></returns>
        public SinglePayment this[int index]
        {
            get { return _payments[index]; }
            set { _payments[index] = value; }
        }
        public int Count
        {
            get { return _payments.Count; }
        }
        public IEnumerator GetEnumerator()
        {
            return _payments.GetEnumerator();
        }
        #endregion

        #region Public c'tors
        public AmortizationModel()
        {
            init();
        }
        public AmortizationModel(double principal, double apr, PaymentsPeriod paymentsPerYear,
            int qtyPayments, DateTime startDate)
        {
            _principal = principal;
            _apr = apr / 100;
            _paymentsPerYear = paymentsPerYear;
            _qtyPayments = qtyPayments;
            _startDate = startDate;

            init();
        }
        public AmortizationModel(AmortizationBluePrint bluePrint)
            : this(bluePrint.Principal, bluePrint.Apr, (PaymentsPeriod)bluePrint.PaymentsPerYear,
            bluePrint.QtyPayments, bluePrint.StartDate)
        {
            // populate extra payments
            foreach (ExtraPaymentItem item in bluePrint.ExtraPayments)
                this[item.Index].ExtraPayment = item.Payment;
        }
        private void init()
        {
            calculate();

            // subscribe to events
            _payments.Added += new IntEventHandler(_payments_Added);
            _payments.Removed += new IntEventHandler(_payments_Removed);
            _payments.Modified += new IntEventHandler(_payments_Modified);
        }
        #endregion

        private void calculate()
        {
            if (_paymentsPerYear <= 0 || _qtyPayments <= 0)
            {
                OnInvalidLoanData();
                return;
            }

            _qtyYears = _qtyPayments / (double)_paymentsPerYear;
            _monthsBetweenPayments = 12 / (int)_paymentsPerYear;

            double oldPaymentAmount = _paymentAmount;
            _paymentAmount = _principal / ((1 - 1 / Math.Pow(1 + _apr / (int)_paymentsPerYear,
                _qtyPayments)) / (_apr / (int)_paymentsPerYear));
            if (!isValid(_paymentAmount))
                OnInvalidLoanData();
            else if (oldPaymentAmount != _paymentAmount)
                OnPaymentChanged();

            refreshAt(0);
        }

        /// <summary>Recalculate starting at a given index</summary>
        /// <param name="index">zero based index</param>
        private void refreshAt(int index)
        {
            if (index < 0)
                throw new Exception("can't be less than 0");

            double endOfPeriodPrinc;
            if (index > 0)
            {
                // compensate for index++ in while loop
                index--;
                endOfPeriodPrinc = _principal;
            }
            else // index == 0
            {
                // mark irrelevant fields with -1. it'd be confusing with all zeros
                SinglePayment lastPayment = new SinglePayment(
                    -1, _startDate.AddMonths(-_monthsBetweenPayments),
                    -1, -1, -1, -1, -1, 0, 0, _principal);

                endOfPeriodPrinc = updatePaymentRow(index, lastPayment);
            }

            while (endOfPeriodPrinc > 0)
            {
                SinglePayment lastPayment = _payments[index];
                index++;
                endOfPeriodPrinc = updatePaymentRow(index, lastPayment);
            }

            // 1 based index
            int position = index + 1;

            // if needed, trim excess rows from _payments
            while (_payments.Count > position)
                _payments.RemoveAt(_payments.Count - 1);

            if (!evalLoanTerms(position))
                OnInvalidLoanData();
            else
            {
                // fire collective events
                OnAdded();
                OnRemoved();
                OnModified();
            }
        }

        private bool evalLoanTerms(int position)
        {
            //_totalInterestPaid;
            double oldTotalInterestPaid = _totalInterestPaid;
            _totalInterestPaid = _payments[Count - 1].CumulativeInterest;
            if (!isValid(_totalInterestPaid))
                return false;

            //_totalPrincipalPaid;
            double oldTotalPrincipalPaid = _totalPrincipalPaid;
            _totalPrincipalPaid = _payments[Count - 1].CumulativePrincipal;
            if (!isValid(_totalPrincipalPaid))
                return false;

            //_totalRepaid;
            double oldTotalRepaid = _totalRepaid;
            _totalRepaid = _totalInterestPaid + _totalPrincipalPaid;
            if (!isValid(_totalRepaid))
                return false;

            //_qtyPaymentsWithExtraPayments;
            int oldExtraQty = _qtyPaymentsWithExtraPayments;
            _qtyPaymentsWithExtraPayments = position;
            if (_qtyPaymentsWithExtraPayments < 1)
                return false;

            if (oldTotalInterestPaid != _totalInterestPaid ||
                oldTotalPrincipalPaid != _totalPrincipalPaid ||
                oldTotalRepaid != _totalRepaid ||
                oldExtraQty != _qtyPaymentsWithExtraPayments)
            {
                OnLoanTermsChanged();
                return true;
            }

            return true;
        }

        private bool isValid(double currency)
        {
            return !(
                double.IsNaN(currency) ||
                double.IsInfinity(currency) ||
                double.IsNegativeInfinity(currency) ||
                currency <= 0);
        }

        private double updatePaymentRow(int index, SinglePayment lastPayment)
        {
            int position = index + 1;

            if (_payments.Count < position)
                _payments.Add(null);

            DateTime paymentDate = lastPayment.PaymentDate.AddMonths(_monthsBetweenPayments);
            double beginningOfPeriodPrinc = lastPayment.PrincipalAtEndOfPeriod;
            // part of the payment that is interest
            double interest = beginningOfPeriodPrinc * _apr / (int)_paymentsPerYear;
            // part of the payment that is principal
            double principal = Math.Min(beginningOfPeriodPrinc, _paymentAmount - interest);
            double paymentAmount = interest + principal;
            double extraPayment = (_payments[index] == null)
                    ? 0
                    : _payments[index].ExtraPayment;
            double endOfPeriodPrinc = beginningOfPeriodPrinc - principal - extraPayment;
            // don't allow final extra payment to be too much
            if (endOfPeriodPrinc < 0)
                extraPayment = beginningOfPeriodPrinc - paymentAmount + interest;

            double cumInterest = lastPayment.CumulativeInterest + interest;
            double cumPrincipal = lastPayment.CumulativePrincipal + principal + extraPayment;

            if (Math.Round(endOfPeriodPrinc, 2) < 0.01)
                endOfPeriodPrinc = 0;

            SinglePayment payment = new SinglePayment(position, paymentDate,
                beginningOfPeriodPrinc, interest, principal, paymentAmount, extraPayment,
                cumInterest, cumPrincipal, endOfPeriodPrinc);
            payment.ExtraPaymentSet += new IntEventHandler(payment_ExtraPaymentSet);
            _payments[index] = payment;
            return endOfPeriodPrinc;
        }

        public void SetBluePrint(AmortizationBluePrint bluePrint)
        {
            _principal = bluePrint.Principal;
            _apr = bluePrint.Apr / 100;
            _paymentsPerYear = (PaymentsPeriod)bluePrint.PaymentsPerYear;
            _qtyPayments = bluePrint.QtyPayments;
            _startDate = bluePrint.StartDate;

            calculate();

            // populate extra payments
            foreach (ExtraPaymentItem item in bluePrint.ExtraPayments)
                this[item.Index].ExtraPayment = item.Payment;
        }
        public AmortizationBluePrint GetBluePrint()
        {
            return new AmortizationBluePrint(this);
        }

        List<int> addedEventCache = new List<int>();
        List<int> removedEventCache = new List<int>();
        List<int> modifiedEventCache = new List<int>();

        private void _payments_Added(object sender, IntEventArgs e)
        {
            addedEventCache.Add(e.Number);
        }

        private void _payments_Removed(object sender, IntEventArgs e)
        {
            removedEventCache.Add(e.Number);
        }

        private void _payments_Modified(object sender, IntEventArgs e)
        {
            modifiedEventCache.Add(e.Number);
        }

        private void payment_ExtraPaymentSet(object sender, IntEventArgs e)
        {
            // event returns the 1 based index. convert to 0 based
            refreshAt(e.Number - 1);
        }

        #region Events
        protected virtual void OnPaymentChanged()
        {
            if (PaymentChanged != null)
                PaymentChanged(this, new EventArgs());
        }
        public event EventHandler PaymentChanged;

        protected virtual void OnLoanTermsChanged()
        {
            if (LoanTermsChanged != null)
                LoanTermsChanged(this, new EventArgs());
        }
        public event EventHandler LoanTermsChanged;

        protected virtual void OnInvalidLoanData()
        {
            if (InvalidLoanData != null)
                InvalidLoanData(this, new EventArgs());
        }
        public event EventHandler InvalidLoanData;


        protected virtual void OnAdded()
        {
            if (addedEventCache.Count == 0)
                return;

            if (Added != null)
                Added(this, intArrayEventHelper(addedEventCache));
        }
        public event IntArrayEventHandler Added;

        protected virtual void OnRemoved()
        {
            if (removedEventCache.Count == 0)
                return;

            if (Removed != null)
                Removed(this, intArrayEventHelper(removedEventCache));
        }
        public event IntArrayEventHandler Removed;

        protected virtual void OnModified()
        {
            if (modifiedEventCache.Count == 0)
                return;

            if (Modified != null)
                Modified(this, intArrayEventHelper(modifiedEventCache));
        }
        public event IntArrayEventHandler Modified;

        private IntArrayEventArgs intArrayEventHelper(List<int> list)
        {
            int[] arr = list.ToArray();
            list.Clear();
            return new IntArrayEventArgs(arr);
        }
        #endregion
    }
    public class SinglePayment
    {
        private int _paymentNumber;
        private DateTime _paymentDate;
        private double _principalAtBeginningOfPeriod;
        private double _interest;
        private double _principal;
        private double _paymentAmount;
        private double _extraPayment;
        private double _cumulativeInterest;
        private double _cumulativePrincipal;
        private double _principalAtEndOfPeriod;

        #region Public Properties
        /// <summary>Get</summary>
        public int PaymentNumber
        {
            get { return _paymentNumber; }
        }
        /// <summary>Get</summary>
        public DateTime PaymentDate
        {
            get { return _paymentDate; }
        }
        /// <summary>Get</summary>
        public double PrincipalAtBeginningOfPeriod
        {
            get { return _principalAtBeginningOfPeriod; }
        }
        /// <summary>Get</summary>
        public double Interest
        {
            get { return _interest; }
        }
        /// <summary>Get</summary>
        public double Principal
        {
            get { return _principal; }
        }
        /// <summary>Get</summary>
        public double PaymentAmount
        {
            get { return _paymentAmount; }
        }
        /// <summary>Getor set</summary>
        public double ExtraPayment
        {
            get { return _extraPayment; }
            set { SetExtraPayment(value); }
        }
        /// <summary>Get</summary>
        public double CumulativeInterest
        {
            get { return _cumulativeInterest; }
        }
        /// <summary>Get</summary>
        public double CumulativePrincipal
        {
            get { return _cumulativePrincipal; }
        }
        /// <summary>Get</summary>
        public double PrincipalAtEndOfPeriod
        {
            get { return _principalAtEndOfPeriod; }
        }
        #endregion

        public SinglePayment(int paymentNumber, DateTime paymentDate, double principalAtBeginningOfPeriod,
            double interest, double principal, double paymentAmount, double extraPayment,
            double cumulativeInterest, double cumulativePrincipal, double principalAtEndOfPeriod)
        {
            _paymentNumber = paymentNumber;
            _paymentDate = paymentDate;
            _principalAtBeginningOfPeriod = principalAtBeginningOfPeriod;
            _interest = interest;
            _principal = principal;
            _paymentAmount = paymentAmount;
            _extraPayment = extraPayment;
            _cumulativeInterest = cumulativeInterest;
            _cumulativePrincipal = cumulativePrincipal;
            _principalAtEndOfPeriod = principalAtEndOfPeriod;
        }

        public void SetExtraPayment(double payment)
        {
            if (payment == _extraPayment)
                return;

            _extraPayment = payment;

            // call event
            OnExtraPaymentSet();
        }

        #region Events
        public event IntEventHandler ExtraPaymentSet;
        protected virtual void OnExtraPaymentSet()
        {
            if (ExtraPaymentSet != null)
                ExtraPaymentSet(this, new IntEventArgs(_paymentNumber));
        }
        #region // Thread safe version

        // nothing multi-threaded yet, so this is overkill.
        // when thread safety is needed: use code below

        //protected readonly object _eventLock = new object();
        ///// <summary>Delegate backing the ExtraPaymentSet event.</summary>
        //protected IntEventHandler _extraPaymentSet;
        ///// <summary>Event fired when an extra payment is set.</summary>
        //public event IntEventHandler ExtraPaymentSet
        //{
        //    add
        //    {
        //        lock (_eventLock)
        //            _extraPaymentSet += value;
        //    }
        //    remove
        //    {
        //        lock (_eventLock)
        //            _extraPaymentSet -= value;
        //    }
        //}
        //protected virtual void OnExtraPaymentSet()
        //{
        //    eventLauncher(_extraPaymentSet);
        //}
        ///// <summary>Safely launch event.</summary>
        ///// <param name="myEvent">Event to fire.</param>
        //private void eventLauncher(IntEventHandler myEvent)
        //{
        //    IntEventHandler handler;
        //    lock (_eventLock)
        //        handler = myEvent;

        //    if (handler != null)
        //        handler(this, new IntEventArgs(_paymentNumber));
        //}
        #endregion
        #endregion
    }
    public class IntEventArgs : EventArgs
    {
        public int Number;
        public IntEventArgs(int number)
        {
            Number = number;
        }
    }
    public delegate void IntEventHandler(object sender, IntEventArgs e);
    public class IntArrayEventArgs : EventArgs
    {
        public int[] Numbers;
        public IntArrayEventArgs(int[] numbers)
        {
            Numbers = numbers;
        }
    }
    public delegate void IntArrayEventHandler(object sender, IntArrayEventArgs e);
}
